/* 
 * Wisapi.cpp
 *
 * Copyright 1997 Sybase, Inc.  All rights reserved
 */

/*************************************************************************
 *
 * Microsoft Internet Information Server extension code
 * Class WHttpConnectISAPI and exported DLL entry points
 *
 *
 *************************************************************************/

#include <wpch.hpp>
#include "WebService.hpp"
#include <stdio.h>		// _vbprintf()


// How to use this template:
// To build an ISAPI DLL called foo.dll, create a new Optima++ target
// of that name, using the ISAPI Target Type.
// In the WHttpConnect::Service() function, add your code to emit HTML
// or whatever you want to output

// The DLL should be built *without* the Optima++ wclass runtime DLL,
// to avoid runtime linkage problems on systems without WCLASS

// To do this, open the Target View window, select your target
// DLL, and use the Right mouse to open the Properties sheet.
// Under the ExportAliases tab, fill in the aliases (using Add)
// if they are not already defined.
// Under the ExecutableSize tab, turn *off* the "Use the runtime DLL"


// IIS will load the DLL when its filename is referenced as a URL
// from a client browser. When first loaded, GetExtensionVersion is called,
// and must succeed or the DLL is unloaded.

// You may also want to change the lpszExtensionDesc[] field in the
// GetExtensionVersion call. It is not clear if this matters.

/********
 * ISAPI Required Entry Points
 ********/

// GetExtensionVersion
//
// Initial entry point
// Called by web server when this DLL is loaded

extern "C" int WDLLEXPORT GetExtensionVersion (HSE_VERSION_INFO *vp)
{
	vp->dwExtensionVersion = (HSE_VERSION_MAJOR << 16) |
		HSE_VERSION_MINOR;
	(void) strcpy(vp->lpszExtensionDesc, "powerppISAPI");
	return 1;
}


// TerminateExtension
//
// Termination-Request point, used by IIS V3.x and up
// If Flags is set to HSE_TERM_ADVISORY_UNLOAD, your override of this function 
// can return TRUE to indicate that it can safely exit or FALSE to indicate 
// that it is not ready to unload. If Flags is set to HSE_TERM_MUST_UNLOAD, 
// the server is forcing the extension out of memory and the return value is ignored.
//
// Here the Webserver asks if it can unload the ISAPI DLL
//      returning TRUE allows the webserver to unload the DLL
//      returning FALSE prevents the webserver to unload the DLL
#define HSE_TERM_ADVISORY_UNLOAD                   0x00000001   
//
// The server informs the ISAPI DLL that it will be unloaded NOW
//      No way to prevent the unload in this place, just cleanup 
//
#define HSE_TERM_MUST_UNLOAD                       0x00000002

extern "C" BOOL WDLLEXPORT TerminateExtension( WULong flags)
{
    if (flags & HSE_TERM_ADVISORY_UNLOAD)
    {
        WebService conn;

    	if (! conn.Create())
            return true;
            
        conn.SetResultCode(1); // Initialized to allow unload
       	(void) conn.CallEventHandler(WDeactivateEvent, &conn);

	    int retcode = conn.GetResultCode();
	    (void) conn.CallEventHandler(WHttpConnectDestroyEvent, NULL);
	    return retcode;
    }
    else
    {
        if (flags &  HSE_TERM_MUST_UNLOAD)
        {
            WebService conn;
    
            if (! conn.Create())
                return true;
                
            conn.SetResultCode(1); // Initialized to allow unload
            (void) conn.CallEventHandler(WEndEvent, &conn);
    
            int retcode = conn.GetResultCode();
            (void) conn.CallEventHandler(WHttpConnectDestroyEvent, NULL);
            return retcode;
        }
        else
        {
            // ? Why has this method been called by the webserver ?????????????
        }
    }
	return true;
}


// HttpExtensionProc
//
// Called by web server for each request to this DLL
// It creates a new WHttpConnectISAPI and invoke its Service() method

extern "C" WULong WDLLEXPORT HttpExtensionProc (EXTENSION_CONTROL_BLOCK *pECB)
{
	WebService conn;

    //EventLog    ev;
    
    //ev.SetSourceName("aarShopISA");
    //ev.LogInfo("HttpExtensionProc called");

	if (! conn.Create())
        return HSE_STATUS_ERROR;

    (void) conn.SetExtensionBlock(pECB);   // Link ECB to WHttpConnectISAPI

	// Actually do the request processing
	(void) conn.SetResultCode(HSE_STATUS_SUCCESS);
	(void) conn.CallEventHandler(WHttpConnectServiceEvent, &conn);

	int retcode = conn.GetResultCode();

	(void) conn.CallEventHandler(WHttpConnectDestroyEvent, &conn);

	return retcode;
}

// WHttpConnectISAPI methods ----------------


// GetServerVariable
//
// Get the value of a named Web server variable
WString WHttpConnectISAPI::GetServerVariable (const WChar* name)
{
	char tempbuf[400];
	WULong len = sizeof tempbuf;
	char * buf = tempbuf;
	len = sizeof tempbuf;

	if (! _ecb->GetServerVariable(_ecb->ConnID, (char*) name,
			buf, &len))
		return WString::GetNullString();
	WString result = buf;
	return result;
}

// GetIsPostMethod
//
// Determine if this is a posted form,
// using METHOD=POST
WBool WHttpConnectISAPI::GetIsPostMethod()
{
	WString method = GetServerVariable("REQUEST_METHOD");
	if (method.GetNull() || method != WTEXT("POST"))
		return FALSE;
	return TRUE;
}

// ISAPI provides POST and GET data in seperate buffers.
// Parse the buffer, turning & into nul characters,
// and creating name/variable pointers for all entries
//
// WEIRDNESS: there seems to be a newline at the end of the ISAPI POST,
// so a trailing nul can be inserted there, to terminate the last item
//
// For HTML FORM method=GET, the lpszQueryString contains QUERY_STRING values
// which have the same format as POST data

void WHttpConnectISAPI::ParseQueryData()
{
	char * cp = (char*) _ecb->lpszQueryString;
	char * limit = cp + strlen(cp);

	ParseCommonData(FALSE, cp, limit);
}

void WHttpConnectISAPI::ParseFormData()
{
	char * cp = (char *) _ecb->lpbData;
	char * limit = cp + _ecb->cbAvailable;

	ParseCommonData(TRUE, cp, limit);
}

// Get a client HTTP header
// ISAPI mangles the names of HTTP variables
// Lookup the name with HTTP_ prefix, then without;
// since some Web Servers (OReilly) do not provide HTTP_COOKIE, only COOKIE!
WString WHttpConnectISAPI::GetHTTPHeader(const WChar* name)
{
	char httpname[200];
#define	HTTP_PREFIX	"HTTP_"
#define LEN_HTTP_PREFIX	5

	(void) strcpy(httpname, HTTP_PREFIX);
	char * cp = strchr(httpname, 0);
	for (char * sp = (char*) name; *sp; sp++) {
		if (*sp == '-')
			*cp++ = '_';
		else
			*cp++ = toupper(*sp);
	}
	*cp = '\0';
	WString ret = GetServerVariable(httpname);
	if (ret.GetNull())
		ret = GetServerVariable(&httpname[LEN_HTTP_PREFIX]);
	return ret;
}

// Response Headers
// Manipulate a list of response headers, and send them
// HTTP allows a number of response headers to be output
// Header() outputs any saved headers

WString WHttpConnectISAPI::GetResponseHeader(const WChar* name)
{
	WString result = LookupVariable(_headers, name, 0);
	return result;		// force a copy from WConstantString
}

WBool WHttpConnectISAPI::SetResponseHeader(const WChar* name,
	const WChar* value)
{
	int i;
	for (i = 0; i < _headers.GetCount(); i++)
		if (_headers[i].GetName() == name)
			return _headers[i].SetValue(value);

	// Allocate new space in array
	i = _headers.GetCount();
	if (! _headers.SetCount(i+1))
		return FALSE;

	_headers[i] = WHttpNameValue(name, value);
	return TRUE;
}

WBool WHttpConnectISAPI::RemoveResponseHeader(const WChar* name)
{
	int cnt = _headers.GetCount();
	for (WInt i = 0; i < cnt; i++) {
		if (_headers[i].GetName() != name)
			continue;
		if (i+1 < cnt) {
			// Since order does not matter,
			// copy last item on top of ith
			_headers[i] = _headers[cnt-1];
		}
		_headers.SetCount(cnt-1);
 /**/		return TRUE;
	}
	return FALSE;
}

// Redirect the client to another URL, by sending a 302 URL Redirect message
// Effectively end this transaction
WBool WHttpConnectISAPI::Redirect(const WString& newurl)
{
	char * url = (char*) newurl.GetAnsiText();
	WULong len = strlen(url);
	return _ecb->ServerSupportFunction(_ecb->ConnID,
		HSE_REQ_SEND_URL_REDIRECT_RESP,
		url,
		&len,
		0L);
}


// Log
//
// Format a log message into ISAPI _ecb
// Subsequent calls overwrite the same log message within a request
void WHttpConnectISAPI::Log(int level, const WChar * where,
	const WChar * fmt, ...)
{
	va_list	arglist;

	va_start(arglist, fmt);
	_vbprintf( _ecb->lpszLogData, HSE_LOG_BUFFER_LEN,
		fmt, arglist);
	va_end(arglist);

	// Ensure nul-terminated string
	*(_ecb->lpszLogData + HSE_LOG_BUFFER_LEN - 1) = '\0';
}

// Low-level read function. There have been reports that no matter
// how large a POST comes in, the server sucks it all into virtual memory
// and provides a handle. So the extra work here is moot..
WInt WHttpConnectISAPI::Read(void * buf, WInt blen)
{
	if (_offset >= _ecb->cbTotalBytes)	// no more data
		return (-1);
	WULong cnt;
	if (_offset < _ecb->cbAvailable) {	// in available memory
		cnt = (_offset + blen) - _ecb->cbAvailable;
		if (cnt <= 0)	{ // all in buffer
			(void) memcpy(buf, _ecb->lpbData + _offset, blen);
			_offset += blen;
			return blen;
		}
		cnt = _ecb->cbAvailable - _offset;
		(void) memcpy(buf, _ecb->lpbData + _offset, cnt);
		_offset = _ecb->cbAvailable;
		return cnt;
	} else
		cnt = blen;
	if (! _ecb->ReadClient(_ecb->ConnID, buf, &cnt))
		return FALSE;
	// actually read cnt bytes. If zero, client socket is closed: fail
	if (cnt == 0)
		return -1;
	_offset += cnt;
	return cnt;
}

// Write
//
// Write data to browser client
// If no header has been output, force out HTTP headers
WBool WHttpConnectISAPI::Write(const void * buf, WInt blen)
{
	if (! _responded)
		(void) SendHeaders(200);

	WULong cnt = blen;
	if (! _ecb->WriteClient(_ecb->ConnID, (void *) buf, &cnt, 0))
		return FALSE;
	if (cnt != blen)
		return FALSE;
	return TRUE;
}

// Low level response: send any queued HTTP headers, along with
// a response code and optional message (default is OK)
// Must be sent before any real output, or as part of returning an error
// If no Content-type has been sent yet, force to text/html
WBool WHttpConnectISAPI::SendHeaders(int retcode, const WChar * msg)
{
	if (_responded)			// already responded, sorry..
		return FALSE;

	WString ctype = GetResponseHeader("Content-Type");
	if (ctype.GetNull()) {
		ctype = WCTEXT("text/html");
		(void) SetResponseHeader("Content-Type", ctype);
	}

	char	temp[4096];		// Build response buffer..
	char	* cp = temp;
	WULong	len = sizeof temp - 1;

	*cp = '\0';
	for (WInt i = 0; i < _headers.GetCount(); i++) {
		(void) _bprintf(cp, len, "%s: %s\n",
			_headers[i].GetName().GetAnsiText(),
			_headers[i].GetValue().GetAnsiText());
		cp = strchr(cp, '\0');
		len = sizeof temp - (cp - temp);
		if (len < 20)
			break;
	}
	(void) strcpy(cp, "\r\n");

	_responded = TRUE;

	char	shortcode[250];
	if (! msg)
		msg = "OK";
	(void) sprintf(shortcode, "%d %s", retcode, msg);

	len = strlen(temp) + 1;
	if (! _ecb->ServerSupportFunction ( _ecb->ConnID,
			HSE_REQ_SEND_RESPONSE_HEADER,
			shortcode,
			&len,
			(WULong *) temp) )
		return FALSE;

	// Output initial <html> tag
	if ( ctype == WCTEXT("text/html") ) {
		ctype = WCTEXT("<HTML>");
		return Write( ctype.GetAnsiText(), ctype.GetAnsiSize() );
	} else
		return TRUE;
}
